#!/usr/bin/perl
use strict;
use warnings;
use POSIX qw(ceil floor);

# Usage: ./driver (parsed_data_file)

# The input data file assumes that each line corresponds to measurements
# of a each metric. So if we have n lines we have n metrics.

require "config.pl";

############################################################################
# Control variables
############################################################################
our $num_metrics;
our $window_size;
our $max_p_order;
our $max_d_order;
our $max_q_order;
our $r_path;
############################################################################

print "Configuration: num_metrics=$num_metrics, win_size=$window_size, (max) [q=$max_p_order, d=$max_d_order, q=$max_q_order]\n";

#### Get input data
my $file_name = $ARGV[0];
my @data;

## Load data from parsed file ###
print "Load data from parsed file...\n";

my @lines;
open FILE, $file_name or die $!;
while (my $line = <FILE>) {
	push(@lines, $line);
}
close(FILE);

my $num_lines = @lines;
my $samples = 0;

for (my $i=0; $i < $num_metrics; $i++) {
	my @tmp = split(/,/, $lines[$i]);
	push(@data, \@tmp);
	
	$samples = @tmp;
}

print "Num. Samples = $samples, Num_lines = $num_lines, Num_metrics = $num_metrics\n";

##################################
# Train model for each metric
##################################

# Maximum value to eliminate buggy observations > e+10
my $maxValue = 1e+10;

my %model;
my %aic;
my %params;
my $num_windows = floor($samples/$window_size);
printf "Number of windows = $num_windows\n";

system "rm -f ./windows/*";

my $path_win = "./windows";
printf "Training models...(THIS WILL TAKE A WHILE)\n";
for (my $metric=0; $metric < @data; $metric++) {
	printf "Metric $metric\n";
	for (my $w=1; $w <= $num_windows; $w++) {
		
		### Create data files for each 
		### window, for each metric 
		my $m = $metric + 1;
		my $file_name = $path_win."/win_"."$m"."_"."$w";
		open FILE, ">", $file_name or die $!;
			my $start = $w * $window_size - $window_size;
			my $stop = $start + $window_size - 1;
			for (my $k=$start; $k <= $stop; $k++) {
				# Eliminate buggy values in data (e.g. e+018 values)
				if ($data[$metric][$k] > $maxValue) {
					$data[$metric][$k] = 0;
				}
				print FILE "$data[$metric][$k]\n";		
			}
		close(FILE);
		
		############# Create models ############################
		# Loop for multiple models ####
		for (my $p=0; $p <= $max_p_order; $p++) {
			for (my $d=0; $d <= $max_d_order; $d++) {
				for (my $q=0; $q <= $max_q_order; $q++) {
					
					#printf "Metric: $metric, w=$w, p=$p, d=$d, q=$q\n";
			
					### Model we don't want to evaluate:
					### ARIMA(0,0,0)
					### ARIMA(0,1,0)
					### ARIMA(0,2,0)
					if (($p==0 && $d==0 && $q==0) || ($p==0 && $d==1 && $q==0) || ($p==0 && $d==2 && $q==0) || ($p==0 && $d==3 && $q==0))  {
						next;
					}

					### Create the script and close the file ###
					#printf "Creating R script\n";

					open FILE, ">training_script.R" or die $!;
					print FILE "#!$r_path\n";
					print FILE "library(fArma)\n";
					print FILE "x <- read.table(\"$file_name\", header=FALSE)\n";
					print FILE "fit = armaFit(~ arima($p, $d, $q), data = x)\n";
					print FILE "summary(fit)\n";
					print FILE "q()\n";

					close FILE;
	
					system("chmod +x training_script.R");
	
					### Run the script ###
					#printf "MODEL: ARIMA($p, $d, $q)\n";
					my @line = `./training_script.R 2> /dev/null`;
					for (my $i=0; $i < @line; $i++) {
						if ($line[$i] =~ m/^Model:/) {
							$model{$p}{$d}{$q}{$metric}{$w} = $line[$i+1];
							next;
						}
	
						if ($line[$i] =~ m/^AIC Criterion/) {
							my @tmp = split(/\s+/, $line[$i]);
							$aic{$p}{$d}{$q}{$metric}{$w} = $tmp[2];
						}
					
						if ($line[$i] =~ m/^Coefficient/) {
							if ($line[$i+4] =~ m/^Residuals/) {
								my @tok = split(/\s+/, $line[$i+2]);
								my $tmp = "";
								if ($tok[0] eq "") {
									#printf "Entered--------\n";
									for (my $y=1; $y < @tok; $y++) {
										$tmp = $tmp."$tok[$y]";
										if ($y != (@tok - 1)) {
											$tmp = $tmp."|";
										} 
									}
								} else {
									for (my $y=0; $y < @tok; $y++) {
										$tmp = $tmp."$tok[$y]";
										if ($y != (@tok - 1)) {
											$tmp = $tmp."|";
										} 
									}
								}
								$params{$p}{$d}{$q}{$metric}{$w} = $tmp;
								#last;
							}
						}
					}
					#printf "params=$params{$p}{$d}{$q}{$metric}{$w}\n";
				}
			}
		}
		###### Finish with models ###################################
	}
}


#####################################
# Select best model for each metric
#####################################
printf "Selecting the best model...\n";
my $path_models = "./models";
system "rm -f ./models/*";
for (my $metric=0; $metric < @data; $metric++) {
	my $m = $metric + 1;
	my $file_metrics = $path_models."/metric_"."$m";
	open FILE, ">$file_metrics" or die $!;
	for (my $w=1; $w <= $num_windows; $w++) {
		### Look for the model with the best AIC ###
		my $min_aic = 100000000;
		my $best_p = -1;
		my $best_d = -1;
		my $best_q = -1;
		for (my $p=0; $p <= $max_p_order; $p++) {
        	for (my $d=0; $d <= $max_d_order; $d++) {
                for (my $q=0; $q <= $max_q_order; $q++) {
                	
                	### Model we don't want to evaluate:
					### ARIMA(0,0,0)
					### ARIMA(0,1,0)
					### ARIMA(0,2,0)
					if (($p==0 && $d==0 && $q==0) || ($p==0 && $d==1 && $q==0) || ($p==0 && $d==2 && $q==0))  {
						next;
					}

					if (!$aic{$p}{$d}{$q}{$metric}{$w}) {
						next;
					}

					if ($aic{$p}{$d}{$q}{$metric}{$w} < $min_aic) {
						$min_aic = $aic{$p}{$d}{$q}{$metric}{$w};
						$best_p = $p;
						$best_d = $d;
						$best_q = $q;
					}
				}
			}
		}
		
		#### Save models in files ####
		
		if ($best_p == -1 && $best_d == -1 && $best_q == -1) {
			print FILE "0/0/0|\n";
			#printf "Metric: $metric, w=$w, Params=0/0/0\n";
		} else {
			my @par_values = split(/\|/, $params{$best_p}{$best_d}{$best_q}{$metric}{$w});
			print FILE "$best_p/$best_d/$best_q|";
			for (my $x=0; $x < @par_values; $x++) {
				print FILE "$par_values[$x]";
				if ($x != (@par_values - 1)) {
					print FILE "/";
				}
			}
			print FILE "\n";
			#printf "Metric: $metric, w=$w, Params: $params{$best_p}{$best_d}{$best_q}{$metric}{$w}\n";
		}
	}
	close(FILE);
}

printf "done!\n";




